// /O3 /QxW /Qipo

//
// A series of parallel pi programs  
// Some OpenMP Commands to support Exercises
// 


static long num_steps = 100000000;
double step;
#define NUM_THREADS 2
  
//  
// PI Program: an example 
//
//#pragma optimize("", off)

void main1() {
  int i;  
  double x, pi;
  double sum = 0.0;

  step = 1.0 / (double)num_steps; 

  for( i = 1; i <= num_steps; i++ ){
    x = (i - 0.5) * step;
    sum = sum + 4.0 / (1.0 + x * x);
  }

  pi = step * sum;
}
#pragma optimize("", on)
 
//  
// Parallel Pi Program 
//
// Lets speed up the program with multiple threads.
//
// Consider the Win32 threads library:
//   - Thread management and interaction is explicit.
//   - Programmer has full control over the threads
//

// 
// Win32 API PI Program
//
#pragma optimize("", off)

#include <windows.h>
HANDLE thread_handles[NUM_THREADS];
CRITICAL_SECTION hUpdateMutex;
double global_sum = 0.0; 

void Pi( void *arg ) {
  int i, start;
  double x, sum = 0.0; 

  start = *(int *) arg;

  step = 1.0 / (double)num_steps; 

  for( i = start; i <= num_steps; i = i + NUM_THREADS ){
    x = (i - 0.5) * step;
    sum = sum + 4.0 / (1.0 + x * x);
  }

  EnterCriticalSection( &hUpdateMutex );
  global_sum += sum;
  LeaveCriticalSection( &hUpdateMutex );
} 

void main2() {
  double pi; int i;
  DWORD threadID;
  int threadArg[NUM_THREADS]; 

  for( i = 0; i < NUM_THREADS; i++)
    threadArg[i] = i + 1; 
    
  InitializeCriticalSection( &hUpdateMutex ); 

  for( i = 0; i < NUM_THREADS; i++ ) {
    thread_handles[i] = CreateThread(0, 0, (LPTHREAD_START_ROUTINE) Pi, &threadArg[i], 0, &threadID );
  } 

  WaitForMultipleObjects( NUM_THREADS, thread_handles, TRUE,INFINITE ); 

  pi = global_sum * step; 

//  printf( " pi is %f \n", pi );
} 

#pragma optimize("", on) 
      
//
// Solution: Keep it simple 
//  
// Threads libraries:
//   Pro: Programmer has control over everything
//   Con: Programmer must control everything 
// Full control 
// Increased complexity 
// Programmers scared away 
//
// Sometimes a simple evolutionary approach is better
//
 
#include <omp.h>

//
// OpenMP PI Program: Parallel Region example (SPMD Program) 
//
void main3() {   
  int i;
  double x, pi, sum[NUM_THREADS] = {0.0};

  step = 1.0 / (double)num_steps;

  omp_set_num_threads( NUM_THREADS );

  #pragma omp parallel

  {   
    double x;
    int i, id;

    id = omp_get_thread_num();

    for( i = id; i < num_steps; i = i + NUM_THREADS ) {
      x = (i + 0.5) * step;
      sum[id] += 4.0 / (1.0 + x * x);
    }
  }

  for( i = 0, pi = 0.0; i < NUM_THREADS; i++ )
    pi += sum[i] * step;
} 

//
// SPMD Programs:
// Each thread runs the same code with the thread ID selecting any thread specific behavior.
//
 
//
// OpenMP PI Program: Work sharing construct 
//
void main4() {  
  int i;
  double x, pi, sum[NUM_THREADS] = {0.0};

  step = 1.0 / (double)num_steps;

  omp_set_num_threads(NUM_THREADS);

  #pragma omp parallel
  
  {   
    double x;
    int i, id;

    id = omp_get_thread_num();     

    #pragma omp for

    for( i = id; i < num_steps; i++ ) {
      x = (i + 0.5) * step;
      sum[id] += 4.0 / (1.0 + x * x);
    }  
  }   
   
  for( i = 0, pi = 0.0; i < NUM_THREADS; i++ )
    pi += sum[i] * step;
}
 
//
// OpenMP PI Program: private clause and a critical section 
//
void main5() {
  int i, id;
  double x, sum, pi=0.0;

  step = 1.0 / (double)num_steps;
  omp_set_num_threads( NUM_THREADS );

  #pragma omp parallel private (x, sum,i)

  { 
    id = omp_get_thread_num();
    
    for( i = id, sum =0.0; i < num_steps; i = i + NUM_THREADS ) {
      x = (i + 0.5) * step;
      sum += 4.0 / (1.0 + x * x);
    }

    #pragma omp critical

    pi += sum * step;
  }
} 
   
//   
// Note: We didnt need to create an array to hold local sums or clutter the code with explicit declarations of x and sum.
// 

//
// OpenMP PI Program: Parallel for with a reduction 
//
void main6() {
  int i;
  double x, pi, sum = 0.0;

  step = 1.0 / (double)num_steps;

  omp_set_num_threads(NUM_THREADS);

  #pragma omp parallel for reduction(+:sum) private(x)

  for( i  = 1; i <= num_steps; i++ ) {
    x = (i - 0.5) * step + x;
   sum = sum + 4.0 / (1.0 + x * x);
  }

  pi = step * sum;
} 
/*
//
// MPI Pi program 
//
#include <mpi.h>

void main7( int argc, char *argv[] ) {
  int i, my_id, numprocs;  
  double x, pi, step, sum = 0.0;

  step = 1.0 / (double)num_steps;

  MPI_Init( &argc, &argv );
  MPI_Comm_Rank( MPI_COMM_WORLD, &my_id );
  MPI_Comm_Size( MPI_COMM_WORLD, &numprocs );

  my_steps = num_steps / numprocs;

  for( i = my_id * my_steps; i < (my_id + 1) * my_steps; i++ ) {
    x = (i + 0.5) * step;
    sum += 4.0 / (1.0 + x * x);
  }

  sum *= step ;

  MPI_Reduce( &sum, &pi, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
}
*/


static double GetPerformanceTime( const LARGE_INTEGER & startCount ) {
  LARGE_INTEGER finishCount;
  QueryPerformanceCounter( &finishCount );

  static LARGE_INTEGER s_freq = {0};
  if( s_freq.QuadPart == 0 )
    QueryPerformanceFrequency( &s_freq );

  return ((double)(finishCount.QuadPart - startCount.QuadPart) / (double)s_freq.QuadPart) * 1000.0;
}

#include <iostream>
using namespace std;

int main(int argc, char* argv[]) {
  LARGE_INTEGER pc;

//  QueryPerformanceCounter( &pc ); Sleep(1000); cout << GetPerformanceTime( pc ) << endl;

  QueryPerformanceCounter( &pc );
  main1();
  cout << "1: " << GetPerformanceTime( pc ) << endl;

//Sleep(10000);

  QueryPerformanceCounter( &pc );
  main2();
  cout << "2: " << GetPerformanceTime( pc ) << endl;

  QueryPerformanceCounter( &pc );
  main3();
  cout << "3: " << GetPerformanceTime( pc ) << endl;

  QueryPerformanceCounter( &pc );
  main4();
  cout << "4: " << GetPerformanceTime( pc ) << endl;

  QueryPerformanceCounter( &pc );
  main5();
  cout << "5: " << GetPerformanceTime( pc ) << endl;

  QueryPerformanceCounter( &pc );
  main6();
  cout << "6: " << GetPerformanceTime( pc ) << endl;

//  QueryPerformanceCounter( &pc ); Sleep(1000); cout << GetPerformanceTime( pc ) << endl;

  system("pause");

	return 0;
}

/* /Qopenmp
2.26146
275.492
276.234
274.878
274.966
276.401
*/

/* /Qopenmp /O3
2.15251
280.864
276.883
275.954
275.05
276.668
*/

/* /Qopenmp /O3 /QxW
1.38509
252.016
253.43
275.176
276.271
275.972
*/

/* /Qopenmp /O3 /QxW /Qipo
0.00279365
252.382
255.436
275.819
275.076
274.88
*/
